# __main__.py

import configparser
from argparse import ArgumentParser
from configparser import ConfigParser
from pathlib import Path
from sys import argv

from . import utils, maxmind, dbip


def get_version():
    root_dir = Path(__file__).parent
    return (root_dir / 'VERSION').read_text()


def get_config_parser(path):
    """
    Returns a ConfigParser, or None.
    """
    try:
        config_file = ConfigParser(allow_no_value=True)
        config_file.read(path)
        return config_file
    except configparser.Error as e:
        # TODO: catch each error type separately
        print("Problem loading config file. Ignoring...", e.message)
        return None


def get_config(cli_args=None):
    """
    Generate configuration
    """
    default_config_path = "/etc/geoipsets.conf"
    default_output_dir = "/tmp"

    # to simplify unit testing
    if not cli_args:
        cli_args = argv[1:]

    parser = ArgumentParser(prog="geoipsets",
                            description="""Utility to build country specific IP sets for ipset/iptables and nftables.
                            Command line arguments take precedence over those in the configuration file.""")
    parser.add_argument("-v", "--version",
                        action="version",
                        version="%(prog)s {0}".format(get_version()))
    parser.add_argument("-p", "--provider",
                        action="extend",
                        nargs="+",
                        type=str.lower,
                        choices={'dbip', 'maxmind'},
                        help="dataset provider(s) (default: {0})".format('dbip'))
    parser.add_argument("-f", "--firewall",
                        action="extend",
                        nargs="+",
                        type=str.lower,
                        choices={utils.Firewall.NF_TABLES.value, utils.Firewall.IP_TABLES.value},
                        help="firewall(s) to build sets for (default: {0})".format(utils.Firewall.NF_TABLES.value))
    parser.add_argument("-a", "--address-family",
                        action="extend",
                        nargs="+",
                        type=str.lower,
                        choices={utils.AddressFamily.IPV4.value, utils.AddressFamily.IPV6.value},
                        help="IP protocol(s) to build sets for (default: {0})".format(utils.AddressFamily.IPV4.value))
    parser.add_argument("-l", "--countries",
                        type=str,
                        help="""Path to a file containing 2-character country codes, one per line, or a comma-separated
                             list of country codes. Argument is treated as a path first. If it does not resolve, or
                             the resolved file is invalid, then it is parsed as a comma-separated list.""")
    parser.add_argument("-o", "--output-dir",
                        type=str,
                        help="directory where geoipsets should be saved (default: {0})".format(default_output_dir))
    parser.add_argument("-c", "--config-file",
                        type=str,
                        default=default_config_path,
                        help="path to configuration file (default: {0})".format(default_config_path))
    parser.add_argument("--checksum",
                        dest="checksum",
                        action="store_true",
                        help="enable checksum validation of downloaded files (default)")
    parser.add_argument("--no-checksum",
                        dest="checksum",
                        action="store_false",
                        help="disable checksum validation of downloaded files")
    parser.set_defaults(checksum=True)

    # set defaults
    default_options = dict()
    default_options['general'] = {''}
    default_options['output-dir'] = default_output_dir
    default_options['provider'] = {'dbip'}
    default_options['firewall'] = {utils.Firewall.NF_TABLES.value}
    default_options['address-family'] = {utils.AddressFamily.IPV4.value}
    default_options['countries'] = 'all'
    default_options['checksum'] = parser.parse_args(cli_args).checksum
    options = default_options

    # step 1: load a valid configuration file, if one exists
    config_file = None
    valid_conf_file = True
    if (config_file_path := parser.parse_args(cli_args).config_file) is not None:
        if (config_file := get_config_parser(config_file_path)) is None:
            valid_conf_file = False

    if valid_conf_file and config_file.has_section('general'):
        general = config_file['general']
    else:
        valid_conf_file = False

    # step 2: output_dir
    if (output_dir := parser.parse_args(cli_args).output_dir) is not None:
        options['output-dir'] = output_dir
    else:
        if valid_conf_file and (output_dir := general.get('output-dir')) is not None:
            options['output-dir'] = output_dir

    # step 3: provider
    if (providers := parser.parse_args(cli_args).provider) is not None:
        options['provider'] = set(providers)
    else:
        if valid_conf_file and (providers := general.get('provider')) is not None:
            options['provider'] = set(providers.split(','))

    # step 4: firewall
    if (firewalls := parser.parse_args(cli_args).firewall) is not None:
        options['firewall'] = set(firewalls)
    else:
        if valid_conf_file and (firewalls := general.get('firewall')) is not None:
            options['firewall'] = set(firewalls.split(','))

    # step 5: address family
    if (address_family := parser.parse_args(cli_args).address_family) is not None:
        options['address-family'] = set(address_family)
    else:
        if valid_conf_file and (address_family := general.get('address-family')) is not None:
            options['address-family'] = set(address_family.split(','))

    # step 6: countries
    if (country_arg := parser.parse_args(cli_args).countries) is not None:
        country_set = set()
        try:
            Path(country_arg).resolve(strict=True)
            with open(country_arg, 'r') as country_file:
                for line in country_file:
                    line = line.strip()
                    if not line.startswith('#'):
                        line = line.split('#')[0].strip()
                        if len(line) == 2 and line.isalpha():
                            country_set.add(line.lower())
        except FileNotFoundError:
            print("file '{0}' does not exist, parsing as list instead".format(country_arg))
            for c in country_arg.split(','):
                if len(c) == 2 and c.isalpha():
                    country_set.add(c.lower())

        if len(country_set) > 0:
            options['countries'] = country_set

    else:
        if valid_conf_file and config_file.has_section('countries'):
            countries = set(config_file['countries'].keys())
            if len(countries) > 0:
                options['countries'] = countries

    # step 7: provider options
    if valid_conf_file:
        for p in options.get('provider'):
            if config_file.has_section(p):
                provider_options = config_file[p]
                options[p] = provider_options

    return options


def main():
    opts = get_config()
    providers = opts.get('provider')
    print("Building geoipsets...")

    if "maxmind" in providers:
        mmp = maxmind.MaxMindProvider(opts.get('firewall'),
                                      opts.get('address-family'),
                                      opts.get('checksum'),
                                      opts.get('countries'),
                                      opts.get('output-dir'),
                                      opts.get('maxmind'))
        mmp.generate()

    if "dbip" in providers:
        dbipp = dbip.DbIpProvider(opts.get('firewall'),
                                  opts.get('address-family'),
                                  opts.get('checksum'),
                                  opts.get('countries'),
                                  opts.get('output-dir'))
        dbipp.generate()


if __name__ == "__main__":
    main()



# utils.py

from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path


class Firewall(Enum):
    IP_TABLES = 'iptables'
    NF_TABLES = 'nftables'


class AddressFamily(Enum):
    IPV4 = 'ipv4'
    IPV6 = 'ipv6'


class AbstractProvider(ABC):
    """Abstract base class providing common functionality for all Provider types."""

    def __init__(self, firewall: set, address_family: set, checksum: bool, countries: set, output_dir: str):
        self.ipv4 = AddressFamily.IPV4.value in address_family
        self.ipv6 = AddressFamily.IPV6.value in address_family
        self.nf_tables = Firewall.NF_TABLES.value in firewall
        self.ip_tables = Firewall.IP_TABLES.value in firewall
        self.checksum = checksum
        self.countries = countries
        self.base_dir = Path(output_dir) / 'geoipsets'

    @abstractmethod
    def generate(self):
        pass



# maxmind.py

import hashlib
import os
import shutil
from csv import DictReader
from io import TextIOWrapper
from pathlib import Path
from tempfile import NamedTemporaryFile
from zipfile import ZipFile

import requests
from requests.auth import HTTPBasicAuth

from . import utils


class MaxMindProvider(utils.AbstractProvider):
    """MaxMind IP range set provider."""

    def __init__(self, firewall: set, address_family: set, checksum: bool, countries: set, output_dir: str,
                 provider_options: dict):
        # 'provider_options' is a ConfigParser Section that can be treated as a dictionary.
        # Use this mechanism to introduce provider-specific options into the configuration file.
        super().__init__(firewall, address_family, checksum, countries, output_dir)

        if not (account_id := provider_options.get('account-id')):
            raise SystemExit("ERROR: Account ID cannot be empty")

        if not (license_key := provider_options.get('license-key')):
            raise SystemExit("ERROR: License key cannot be empty")

        self.auth = HTTPBasicAuth(account_id, license_key)
        self.base_url = 'https://download.maxmind.com/geoip/databases/GeoLite2-Country-CSV/download'

    def generate(self):
        zip_file = self.download()  # comment out for testing

        if self.checksum:
            self.check_checksum(zip_file)

        with ZipFile(Path(zip_file.name), 'r') as zip_ref:
            # with ZipFile(Path("/tmp/tmp23pn2bw0.zip"), 'r') as zip_ref:  # replace line above with this for testing

            zip_dir_prefix = os.path.commonprefix(zip_ref.namelist())
            id_cc_map = self.build_id_cc_map(zip_ref, zip_dir_prefix)

            # TODO: run each address-family concurrently?
            if self.ipv4:
                self.build_sets(id_cc_map, zip_ref, zip_dir_prefix, utils.AddressFamily.IPV4)

            if self.ipv6:
                self.build_sets(id_cc_map, zip_ref, zip_dir_prefix, utils.AddressFamily.IPV6)

        zip_file.close()
        os.remove(zip_file.name)

    def build_id_cc_map(self, zip_ref: ZipFile, dir_prefix: str):
        # Build dictionary mapping geoname_ids to ISO country codes
        # {6251999: 'CA', 1269750: 'IN'}
        # example row: 6251999,en,NA,"North America",CA,Canada,0
        #
        # field names:
        # geoname_id, locale_code, continent_code, continent_name, country_iso_code, country_name, is_in_european_union

        locations = 'GeoLite2-Country-Locations-en.csv'
        id_country_code_map = dict()
        with ZipFile(Path(zip_ref.filename), 'r') as zip_file:
            with zip_file.open(dir_prefix + locations, 'r') as csv_file_bytes:
                rows = DictReader(TextIOWrapper(csv_file_bytes))
                for r in rows:
                    if cc := r['country_iso_code']:
                        # configparser forces keys to lower case by default
                        if self.countries == 'all' or cc.lower() in self.countries:
                            id_country_code_map[r['geoname_id']] = cc

        return id_country_code_map

    def build_sets(self, id_country_code_map: dict, zip_ref: ZipFile, dir_prefix: str, addr_fam: utils.AddressFamily):
        # Iterates through IP blocks and builds country-specific IP range lists.
        # field names:
        # network,geoname_id,registered_country_geoname_id,represented_country_geoname_id,is_anonymous_proxy,is_satellite_provider

        ipset_dir = self.base_dir / 'maxmind/ipset' / addr_fam.value
        nftset_dir = self.base_dir / 'maxmind/nftset' / addr_fam.value
        if addr_fam == utils.AddressFamily.IPV4:
            ip_blocks = 'GeoLite2-Country-Blocks-IPv4.csv'
            inet_family = 'family inet'
        else:  # AddressFamily.IPV6
            ip_blocks = 'GeoLite2-Country-Blocks-IPv6.csv'
            inet_family = 'family inet6'

        # dictionary of subnet lists, indexed by filename
        # filename is CC.address_family -- eg. CA.ipv4
        country_subnets = dict()

        with ZipFile(Path(zip_ref.filename), 'r') as zip_file:
            with zip_file.open(dir_prefix + ip_blocks, 'r') as csv_file_bytes:
                rows = DictReader(TextIOWrapper(csv_file_bytes))
                for r in rows:
                    geo_id = r['geoname_id']
                    if not geo_id:
                        geo_id = r['registered_country_geoname_id']
                    if not geo_id:
                        continue

                    try:
                        cc = id_country_code_map[geo_id]
                    except KeyError:
                        continue  # skip CC if not listed in the config file

                    net = r['network']
                    filename_key = cc + '.' + addr_fam.value

                    if filename_key in country_subnets:  # append
                        country_subnets[filename_key].append(net)
                    else:  # create
                        country_subnets[filename_key] = [net]

        # remove old sets if they exist
        if self.ip_tables:
            if ipset_dir.is_dir():
                shutil.rmtree(ipset_dir)
            ipset_dir.mkdir(parents=True)
        if self.nf_tables:
            if nftset_dir.is_dir():
                shutil.rmtree(nftset_dir)
            nftset_dir.mkdir(parents=True)

        #
        # write data to disk
        #
        for set_name, subnets in country_subnets.items():
            set_name_parts = set_name.split('.')
            country_code = set_name_parts[0]

            # write file headers
            # iptables/ipsets
            if self.ip_tables:
                ipset_file = open(ipset_dir / set_name, 'w')
                maxelem = max(131072, 1 if len(subnets) == 0 else (1 << (len(subnets) - 1).bit_length()))
                ipset_file.write("create {0} hash:net {1} maxelem {2} comment\n".format(set_name,
                                                                                        inet_family,
                                                                                        maxelem))

            # nftables set
            if self.nf_tables:
                nftset_file = open(nftset_dir / set_name, 'w')
                nftset_file.write("define " + set_name + " = {\n")

            # write ranges to file(s)
            for subnet in subnets:
                if self.ip_tables:
                    ipset_file.write("add " + set_name + " " + subnet + " comment " + country_code + "\n")

                if self.nf_tables:
                    nftset_file.write(subnet + ",\n")

            if self.ip_tables:
                ipset_file.close()

            if self.nf_tables:
                nftset_file.write("}\n")
                nftset_file.close()

    def download(self):
        # URL: https://download.maxmind.com/geoip/databases/GeoLite2-Country-CSV/download
        # CSV query string: ?suffix=zip

        # The downloaded filename is available in the 'Content-Disposition' HTTP response header.
        # eg. Content-Disposition: attachment; filename=GeoLite2-Country-CSV_20200922.zip
        file_suffix = 'zip'
        zip_url = self.base_url + '?suffix=' + file_suffix

        # download latest ZIP file
        zip_http_response = requests.get(zip_url, auth=self.auth)
        with NamedTemporaryFile(suffix='.' + file_suffix, delete=False) as zip_file:
            zip_file.write(zip_http_response.content)

        return zip_file

    def download_checksum(self):
        # URL: https://download.maxmind.com/geoip/databases/GeoLite2-Country-CSV/download
        # SHA256 query string: ?suffix=zip.sha256
        file_suffix = 'zip.sha256'
        sha256_url = self.base_url + '?suffix=' + file_suffix
        sha256_http_response = requests.get(sha256_url, auth=self.auth)
        with NamedTemporaryFile(suffix='.' + file_suffix, delete=True) as sha256_file:
            sha256_file.write(sha256_http_response.content)
            sha256_file.seek(0)

            return sha256_file.read().decode('utf-8').split()[0]

    def check_checksum(self, zip_ref):
        expected_sha256sum = self.download_checksum()

        # calculate sha256 hash
        with open(zip_ref.name, 'rb') as raw_zip_file:
            sha256_hash = hashlib.sha256()
            # Read and update hash in 8K chunks
            while chunk := raw_zip_file.read(8192):
                sha256_hash.update(chunk)

            computed_sha256sum = sha256_hash.hexdigest()

        # compare downloaded sha256 hash with computed version
        if expected_sha256sum != computed_sha256sum:
            raise SystemExit("ERROR: Computed zip file digest '{0}' does not match expected value '{1}'".format(
                computed_sha256sum, expected_sha256sum
            ))



# dbip.py

import gzip
import hashlib
import shutil
import os
from csv import DictReader
from datetime import datetime
from io import TextIOWrapper
from ipaddress import ip_address, summarize_address_range
from tempfile import NamedTemporaryFile

import requests
from bs4 import BeautifulSoup

from . import utils


class DbIpProvider(utils.AbstractProvider):
    """ DBIP IP range set provider. """

    def __init__(self, firewall: set, address_family: set, checksum: bool, countries: set, output_dir: str):
        super().__init__(firewall, address_family, checksum, countries, output_dir)

    def generate(self):
        """
        While nftables' set facility accepts both IPv4 and IPv6 IP ranges, ipset only accepts IPv4 IP ranges.
        So, for simplicity we convert all ranges into subnets.

        ip_start, ip_end, country
        """
        gzip_ref = self.download()  # comment out for testing
        # dictionary of subnet lists, indexed by filename
        # filename is CC.address_family -- eg. CA.ipv4
        country_subnets = dict()

        with gzip.GzipFile(gzip_ref, 'rb') as csv_file_bytes:
            # with gzip.GzipFile('/tmp/tmphq4qgkfp.csv.gz', 'rb') as csv_file_bytes:

            # validate checksum of the CSV file (not the GZIP file)
            if self.checksum:
                self.check_checksum(csv_file_bytes)

            rows = DictReader(TextIOWrapper(csv_file_bytes), fieldnames=("ip_start", "ip_end", "country"))
            for r in rows:
                cc = r['country']
                # configparser forces keys to lower case by default
                if cc != 'ZZ' and (self.countries == 'all' or cc.lower() in self.countries):
                    ip_start = ip_address(r['ip_start'])
                    ip_version = ip_start.version
                    if (ip_version == 4 and self.ipv4) or (ip_version == 6 and self.ipv6):
                        inet_suffix = 'ipv' + str(ip_version)
                        filename_key = cc + '.' + inet_suffix
                        ip_end = ip_address(r['ip_end'])
                        if self.ip_tables:  # https://github.com/chr0mag/geoipsets/issues/25
                            subnets = [nets.with_prefixlen for nets in summarize_address_range(ip_start, ip_end)]
                            if filename_key in country_subnets:  # append
                                country_subnets[filename_key].extend(subnets)
                            else:  # create
                                country_subnets[filename_key] = subnets
                        else:  # conversion not required for nftables
                            if ip_start == ip_end:  # nftables disallows intervals with the same start & end
                                ip_range = r['ip_start']
                            else:
                                ip_range = r['ip_start'] + '-' + r['ip_end']
                            if filename_key in country_subnets:  # append
                                country_subnets[filename_key].append(ip_range)
                            else:  # create
                                country_subnets[filename_key] = [ip_range]

        self.build_sets(country_subnets)
        os.remove(gzip_ref)

    def build_sets(self, dict_of_lists):
        ipset_dir = self.base_dir / 'dbip/ipset' / utils.AddressFamily.IPV4.value
        nftset_dir = self.base_dir / 'dbip/nftset' / utils.AddressFamily.IPV4.value
        ip6set_dir = self.base_dir / 'dbip/ipset' / utils.AddressFamily.IPV6.value
        nft6set_dir = self.base_dir / 'dbip/nftset' / utils.AddressFamily.IPV6.value

        # remove old sets if they exist
        if self.ip_tables:
            if self.ipv4:
                if ipset_dir.is_dir():
                    shutil.rmtree(ipset_dir)
                ipset_dir.mkdir(parents=True)

            if self.ipv6:
                if ip6set_dir.is_dir():
                    shutil.rmtree(ip6set_dir)
                ip6set_dir.mkdir(parents=True)

        if self.nf_tables:
            if self.ipv4:
                if nftset_dir.is_dir():
                    shutil.rmtree(nftset_dir)
                nftset_dir.mkdir(parents=True)

            if self.ipv6:
                if nft6set_dir.is_dir():
                    shutil.rmtree(nft6set_dir)
                nft6set_dir.mkdir(parents=True)

        for set_name, subnets in dict_of_lists.items():
            set_name_parts = set_name.split('.')
            country_code = set_name_parts[0]
            ip_version = set_name_parts[1]
            if ip_version == utils.AddressFamily.IPV4.value:
                inet_family = 'family inet'
            else:  # AddressFamily.IPV6
                inet_family = 'family inet6'

            # write file headers
            if self.ip_tables:
                ipset_path = self.base_dir / 'dbip/ipset' / ip_version / set_name
                ipset_file = open(ipset_path, 'w')
                maxelem = max(131072, 1 if len(subnets) == 0 else (1 << (len(subnets) - 1).bit_length()))
                ipset_file.write("create {0} hash:net {1} maxelem {2} comment\n".format(set_name, inet_family, maxelem))

            if self.nf_tables:
                nftset_path = self.base_dir / 'dbip/nftset' / ip_version / set_name
                nftset_file = open(nftset_path, 'w')
                nftset_file.write("define " + set_name + " = {\n")

            # write ranges to file(s)
            for subnet in subnets:
                if self.ip_tables:
                    ipset_file.write("add " + set_name + " " + subnet + " comment " + country_code + "\n")

                if self.nf_tables:
                    nftset_file.write(subnet + ",\n")

            if self.ip_tables:
                ipset_file.close()

            if self.nf_tables:
                nftset_file.write("}\n")
                nftset_file.close()

    def download(self):
        """
        eg. https://download.db-ip.com/free/dbip-country-lite-2020-10.csv.gz
        filename: dbip-country-lite-YYYY-MM.csv.gz
        """
        file_suffix = '.csv.gz'
        url = 'https://download.db-ip.com/free/dbip-country-lite-' + datetime.utcnow().strftime('%Y-%m') + file_suffix

        # download latest GZIP file
        http_response = requests.get(url)
        with NamedTemporaryFile(suffix=file_suffix, delete=False) as gzip_file:
            gzip_file.write(http_response.content)

        return gzip_file.name

    def download_checksum(self):
        webpage = 'https://db-ip.com/db/download/ip-to-country-lite'
        # download sha1sum
        webpage_http_response = requests.get(webpage)

        # the page section we're looking for looks like this:
        # <dl class="card-body">
        #     <dt>Format</dt>
        #     <dd>CSV</dd>
        #     <dt>Release</dt>
        #     <dd>January 2022</dd>
        #     <dt>Supported language(s)</dt>
        #     <dd>English</dd>
        #     <dt>Number of records</dt>
        #     <dd>583,730</dd>
        #     <dt>File size</dt>
        #     <dd>24.7 MB</dd>
        #     <dt>MD5SUM</dt>
        #     <dd class="small">6f58a437323f6bc891a9c8fdef96add3</dd>
        #     <dt>SHA1SUM</dt>
        #     <dd class="small">d663790f368afa00e0ac28f2075299e1e30a5054</dd>
        # </dl>

        soup = BeautifulSoup(webpage_http_response.content, "html.parser")

        # we are using the CSV format, not MMDB
        csv_card_body = soup.find('dd', string="CSV")
        csv_sha1sum_tag = csv_card_body.find_next_siblings('dt', string="SHA1SUM")

        return csv_sha1sum_tag[0].find_next_sibling().string

    def check_checksum(self, csv_file_bytes):
        expected_sha1sum = self.download_checksum()

        # calculate the sha1sum of the downloaded file
        sha1_hash = hashlib.sha1()
        # Read and update hash in 8K chunks
        while chunk := csv_file_bytes.read(8192):
            sha1_hash.update(chunk)

        computed_sha1sum = sha1_hash.hexdigest()

        # reset position to beginning of file now that we're done
        csv_file_bytes.seek(0)

        # compare downloaded sha1 hash with computed version
        if expected_sha1sum != computed_sha1sum:
            raise SystemExit("ERROR: Computed CSV file digest '{0}' does not match expected value '{1}'".format(
                computed_sha1sum, expected_sha1sum
            ))
